@@ -14,7 +14,8 @@ except ImportError:
 from txpostgres import txpostgres
 
 from twisted.trial import unittest
-from twisted.internet import defer, posixbase, reactor
+from twisted.internet import defer, main, posixbase, reactor
+from twisted.python import failure
 
 simple_table_schema = """
 CREATE TABLE simple (
@@ -63,8 +64,8 @@ class PollableThing(object):
         self.NEXT_STATE = psycopg2.extensions.POLL_READ
 
     def poll(self):
-        if self.NEXT_STATE is None:
-            raise Exception("no next state")
+        if isinstance(self.NEXT_STATE, Exception):
+            raise self.NEXT_STATE
         return self.NEXT_STATE
 
     def fileno(self):
@@ -193,10 +194,25 @@ class TxPostgresPollingMixinTestCase(Psycopg2TestCase):
         p._pollable.NEXT_STATE = psycopg2.extensions.POLL_OK
 
         d = p.poll()
-        p.connectionLost(RuntimeError("boom"))
-        p.connectionLost(RuntimeError("bam"))
+        p.connectionLost(failure.Failure(RuntimeError("boom")))
+        p.connectionLost(failure.Failure(RuntimeError("bam")))
         return d.addCallback(self.assertEquals, p)
 
+    def test_connectionLostWhileWaiting(self):
+        """
+        If the connection is lost while waiting for the socket to become
+        writable, the C{Deferred} returned from poll() still errbacks.
+        """
+        p = FakeWrapper()
+        p._pollable = PollableThing()
+        p._pollable.NEXT_STATE = psycopg2.extensions.POLL_WRITE
+
+        d = p.poll()
+        p._pollable.NEXT_STATE = RuntimeError("forced poll error")
+        p.connectionLost(failure.Failure(main.CONNECTION_LOST))
+        return self.assertFailure(d, RuntimeError)
+    test_connectionLostWhileWaiting.timeout = 5
+
     def test_errors(self):
         """
         Unexpected results from poll() make L{txpostgres._PollingMixin} raise
@@ -606,17 +622,36 @@ class TxPostgresQueryTestCase(_SimpleDBSetupMixin, Psycopg2TestCase):
         def terminateAndRunQuery():
             d = self.conn.runQuery("select pg_terminate_backend(%s)",
                                    (self.conn.get_backend_pid(), ))
-            d = self.assertFailure(d, psycopg2.DatabaseError)
-            d.addCallback(lambda _: self.conn.runQuery("select 1"))
-            d = self.assertFailure(d, psycopg2.InterfaceError)
 
-            def restoreConnection():
+            def fail(ignore):
+                self.fail("did not catch an error, instead got %r" % (ignore,))
+
+            def checkDatabaseError(f):
+                if f.check(psycopg2.DatabaseError):
+                    return f.value
+
+                if f.check(SystemError):
+                    raise unittest.SkipTest(
+                        "This test fails on versions of psycopg2 before 2.4.1 "
+                        "which have a bug in libpq error handling")
+
+                self.fail(("\nExpected: %r\nGot:\n%s"
+                           % (psycopg2.DatabaseError, str(f))))
+
+            def runSimpleQuery(_):
+                d = self.conn.runQuery("select 1")
+                return self.assertFailure(d, psycopg2.InterfaceError)
+
+            def restoreConnection(res):
                 self.conn = txpostgres.Connection()
-                return self.conn.connect(user=DB_USER, password=DB_PASS,
-                                         host=DB_HOST, database=DB_NAME)
+                d = self.conn.connect(user=DB_USER, password=DB_PASS,
+                                      host=DB_HOST, database=DB_NAME)
+                return d.addCallback(lambda _: res)
 
+            d.addCallbacks(fail, checkDatabaseError)
+            d.addCallback(runSimpleQuery)
             # restore the connection, otherwise all the other tests will fail
-            return d.addCallback(lambda _: restoreConnection())
+            return d.addBoth(restoreConnection)
 
         return d.addCallback(lambda _: terminateAndRunQuery())
 
@@ -635,7 +670,7 @@ class TxPostgresQueryTestCase(_SimpleDBSetupMixin, Psycopg2TestCase):
         d2 = self.conn.runQuery("select 1")
 
         self.assertEquals(len(cursors), 1)
-        cursors[0].connectionLost(RuntimeError("boom"))
+        cursors[0].connectionLost(failure.Failure(RuntimeError("boom")))
 
         d = defer.gatherResults([d1, d2])
         d.addCallback(self.assertEquals, [[(1, )], [(1, )]])
