@@ -124,7 +124,14 @@ class _PollingMixin(object):
         # correct failure. If we errback here, we won't finish the poll()
         # cycle, which would leave psycopg2 in a state where it thinks there's
         # still an async query underway.
-        return
+        #
+        # If the connection got lost right after the first poll(), the Deferred
+        # returned from it will never fire, leaving the caller hanging forever,
+        # unless we push the connection state forward here. OTOH, if the
+        # connection is already closed, there's no pollable to poll, so check
+        # that before calling poll().
+        if self.pollable():
+            self.poll()
 
     def _cancel(self, d):
         try:
@@ -164,7 +171,7 @@ class Cursor(_PollingMixin):
         self._cursor = cursor
 
     def pollable(self):
-        return self._cursor.connection
+        return self._connection.pollable()
 
     def execute(self, query, params=None):
         """
@@ -206,12 +213,6 @@ class Cursor(_PollingMixin):
         _cursor, self._cursor = self._cursor, None
         return _cursor.close()
 
-    def fileno(self):
-        if self._cursor and self._connection._connection:
-            return self._cursor.connection.fileno()
-        else:
-            return -1
-
     def __getattr__(self, name):
         # the pollable is the connection, but the wrapped object is the cursor
         return getattr(self._cursor, name)
